import subprocess


def install_library(library):
    try:
        result = subprocess.run(
            ["pip", "install", library],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.returncode == 0
    except subprocess.CalledProcessError:
        return False


def uninstall_library(library):
    try:
        result = subprocess.run(
            ["pip", "uninstall", "-y", library],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.returncode == 0
    except subprocess.CalledProcessError:
        return False



try:
    from .utils.db import *

except ImportError:
    from utils.db import *

llm_settings = {
    "gpt-4o": {
        "show_name": "gpt-4o (OpenAI)",
        "vision": True,
        "provider": "openai",
        "tools": True,
        "stream": True,
    },
    "claude-3-5-sonnet-20241022": {
        "show_name": "claude-3-5-sonnet-20241022 (Anthropic)",
        "vision": True,
        "provider": "anthropic",
        "tools": True,
        "stream": False,
    },
    "us.anthropic.claude-3-5-sonnet-20241022-v2:0": {
        "show_name": "claude-3-5-sonnet-20241022 (AWS)",
        "vision": True,
        "provider": "aws",
        "tools": True,
        "stream": False,
    },
    "gpt-4o-azureopenai": {
        "show_name": "gpt-4o (AzureAI)",
        "vision": True,
        "provider": "azureai",
        "tools": True,
        "stream": True,
    },
    "gpt-4o-mini": {
        "show_name": "gpt-4o-mini (OpenAI)",
        "vision": True,
        "provider": "openai",
        "tools": True,
        "stream": True,
    },
    "gpt-4-turbo": {
        "show_name": "gpt-4-turbo (OpenAI)",
        "vision": False,
        "provider": "openai",
        "tools": True,
        "stream": True,
    },
    "gpt-3.5": {
        "show_name": "gpt-3.5 (OpenAI)",
        "vision": False,
        "provider": "openai",
        "tools": True,
        "stream": True,
    },
    "gpt-3.5-turbo": {
        "show_name": "gpt-3.5-turbo (OpenAI)",
        "vision": False,
        "provider": "openai",
        "tools": True,
        "stream": True,
    },
    "llama3": {
        "show_name": "Llama3 (Ollama)",
        "vision": False,
        "provider": "ollama",
        "tools": False,
        "stream": False,
    },
    "llama3.1": {
        "show_name": "Llama3.1 (Ollama)",
        "vision": False,
        "provider": "ollama",
        "tools": True,
        "stream": False,
    },
    "qwen2:1.5b": {
        "show_name": "Qwen2 1.5b (Ollama)",
        "vision": False,
        "provider": "ollama",
        "tools": False,
        "stream": False,
    },
    "llava": {
        "show_name": "Llava (Ollama)",
        "vision": True,
        "provider": "ollama",
        "tools": False,
        "stream": False,
    },
    "bakllava": {
        "show_name": "BakLLaVA (Ollama)",
        "vision": True,
        "provider": "ollama",
        "tools": False,
        "stream": False,
    },
    "llava-llama3": {
        "show_name": "Llava-Llama3 (Ollama)",
        "vision": True,
        "provider": "ollama",
        "tools": False,
        "stream": False,
    },
    "llava-phi3": {
        "show_name": "LLaVA-Phi-3 (Ollama)",
        "vision": True,
        "provider": "ollama",
        "tools": False,
        "stream": False,
    },
    "gemini-pro": {
        "show_name": "gemini-pro (Google)",
        "vision": True,
        "provider": "google",
        "tools": True,
        "stream": True,
    },
    "mixtral-8x7b-groq": {
        "show_name": "Mixtral 8x7b (Groq)",
        "vision": False,
        "provider": "groq",
        "tools": True,
        "stream": True,
    },
}


def get_openai_models():
    return [k for k, v in llm_settings.items() if v["provider"] == "openai"]

def get_azureai_models():
    return [k for k, v in llm_settings.items() if v["provider"] == "azureai"]


def get_ollama_models():
    return [k for k, v in llm_settings.items() if v["provider"] == "ollama"]


def get_google_models():
    return [k for k, v in llm_settings.items() if v["provider"] == "google"]


def get_groq_models():
    return [k for k, v in llm_settings.items() if v["provider"] == "groq"]


llm_show_name_ = {}
for k, v in llm_settings.items():
    llm_show_name_[v["show_name"]] = k

llm_show_name = llm_show_name_


def first_message():
    from .character import name, developer, get_website_content
    from .cu.computer import width, height, display_num
    from .utils.db import load_system_prompt
    model = load_model_settings()

    the_text = f"""
{load_system_prompt()}
"""

    the_website_content = get_website_content()
    if the_website_content:
        the_text += f"""
# The Website Content of the User

{the_website_content}

"""

    return the_text





import json
import re
from langchain.tools import tool
import traceback

import pyautogui
import time

try:
    from .utils.db import *
    from .llm import get_model
    from .top_bar_wrapper import wrapper
    from .llm_settings import llm_settings

except ImportError:
    from utils.db import *
    from top_bar_wrapper import wrapper
    from llm_settings import llm_settings












def mouse_scroll_(direction: str, amount: int = 1) -> bool:
    """
    A function to scroll the mouse wheel.

    Parameters:
    - direction (str): The direction of the scroll. Possible values are "up" and "down".
    - amount (int): The amount of scrolling to be performed. The default value is 1.

    Returns:
    - bool: True if the scrolling was performed successfully, False otherwise.
    """
    try:
        import pyautogui

        pyautogui.FAILSAFE = False

        if direction == "up":
            pyautogui.scroll(amount)
        elif direction == "down":
            pyautogui.scroll(-amount)
        return True
    except:
        traceback.print_exc()
        return False


mouse_scroll = tool(mouse_scroll_)



def extract_code_from_result(llm_output):
    """
    Extract the Python code from the LLM output.
    """
    code_match = re.search(r'```json\n(.*?)```', llm_output, re.DOTALL)
    if code_match:
        return code_match.group(1).strip()
    return llm_output.strip()


def click_to_text_(text:str, double_click=False) -> bool:
    """
    Click on the text
   
    """

    try:
        from .cu.ask_anthropic import ask_anthropic
        from .cu.computer import click_action, mouse_move_action
    except ImportError:
        from cu.ask_anthropic import ask_anthropic
        from cu.computer import click_action, mouse_move_action

    print("click_to_text")
    print("text", text)
    x_y = ask_anthropic(f"dont use tools, give me exactly location of '{text}' text as json x,y like"+ """{'x': 0, 'y': 0}"""+". Only return the json with ```json ```")
    print("result", x_y)

    x_y = extract_code_from_result(x_y)

    x_y = json.loads(x_y)


    pyautogui.click(x_y['x'], x_y['y'], button='left')
    if double_click:
        time.sleep(0.1)
        pyautogui.click(x_y['x'], x_y['y'], button='left')

    return True


click_to_text = tool(click_to_text_)



def click_to_icon_(icon:str, double_click=False) -> bool:
    """
    Click on the icon
   
    """

    try:
        from .cu.ask_anthropic import ask_anthropic
        from .cu.computer import click_action, mouse_move_action
    except ImportError:
        from cu.ask_anthropic import ask_anthropic
        from cu.computer import click_action, mouse_move_action

    print("click_to_icon")
    print("icon", icon)
    x_y = ask_anthropic(f"dont use tools, give me exactly location of '{icon}' icon as json x,y like"+ """{'x': 0, 'y': 0}"""+". Only return the json with ```json ```")
    print("result", x_y)

    x_y = extract_code_from_result(x_y)

    x_y = json.loads(x_y)

    pyautogui.click(x_y['x'], x_y['y'], button='left')
    if double_click:
        time.sleep(0.1)
        pyautogui.click(x_y['x'], x_y['y'], button='left')


    return True


click_to_icon = tool(click_to_icon_)


def click_to_area_(
    area:str, double_click=False
) -> bool:
    """
    Click on the area like search bar
    """

    try:
        from .cu.ask_anthropic import ask_anthropic
        from .cu.computer import click_action, mouse_move_action
    except ImportError:
        from cu.ask_anthropic import ask_anthropic
        from cu.computer import click_action, mouse_move_action

    print("click_to_area")
    print("area", area)
    x_y = ask_anthropic(f"dont use tools, give me exactly location of '{area}' area as json x,y like"+ """{'x': 0, 'y': 0}"""+". Only return the json with ```json ```")
    print("result", x_y)

    x_y = extract_code_from_result(x_y)

    x_y = json.loads(x_y)

    pyautogui.click(x_y['x'], x_y['y'], button='left')
    if double_click:
        time.sleep(0.1)
        pyautogui.click(x_y['x'], x_y['y'], button='left')


    return True




click_to_area = tool(click_to_area_)




def screenshot_(checking:str):
    """
    Returns the current screenshot. Explain what should we check on the screenshot.
    """

    from langchain_core.messages import HumanMessage, SystemMessage, AIMessage

    try:
        from .cu.computer import screenshot_action
        from .agent.agent import get_agent_executor
    except ImportError:
        from cu.computer import screenshot_action
        from agent.agent import get_agent_executor

    the_base64 = screenshot_action(direct_base64=True)






    human_first_message = {"type": "text", "text": f"Explain the image and check '{checking}' on the image."}



    the_message = [
        human_first_message
    ]



    human_second_message = {
            "type": "image_url",
            "image_url": {"url": f"data:image/png;base64,{the_base64}"},
        }




    the_message.append(human_second_message)





    the_message = HumanMessage(content=the_message)







    msg = get_agent_executor(no_tools=True).invoke(
        {"messages": [the_message]}
    )





    the_last_messages = msg["messages"]









    return_value = the_last_messages[-1].content
    if isinstance(return_value, list):
        the_text = ""
        for each in return_value:
            the_text += str(each)
        return_value = the_text

    if return_value == "":
        return_value = "No response "




    return return_value



screenshot = tool(screenshot_)


import random
import traceback


from langchain_core.messages import HumanMessage, SystemMessage, AIMessage




try:

    from ..agent import get_agent_executor
    from ..screen.shot import *
    from ..utils.db import load_model_settings, agents
    from ..llm import get_model
    from ..llm_settings import llm_settings
    from ..utils.chat_history import ChatHistory
    from .computer import screenshot_action
except ImportError:

    from agent import get_agent_executor
    from screen.shot import *
    from utils.db import load_model_settings, agents
    from llm import get_model
    from llm_settings import llm_settings
    from utils.chat_history import ChatHistory
    from computer import screenshot_action






def ask_anthropic(
    the_request:str
):

    try:

        from ..agent import get_agent_executor
    except ImportError:
        from agent import get_agent_executor


    try:
        print("ASK ANTHROPIC", the_request)



        llm_input = the_request

        print("LLM INPUT", llm_input)




        human_first_message = {"type": "text", "text": f"{llm_input}"}



        the_message = [
            human_first_message
        ]



        human_second_message = None



        base64_image = screenshot_action(direct_base64=True)





        human_second_message = {
            "type": "image_url",
            "image_url": {"url": f"data:image/png;base64,{base64_image}"},
        }



        print("LEN OF IMAGE", len(base64_image))


        if human_second_message:
            the_message.append(human_second_message)





        the_message = HumanMessage(content=the_message)


        




        


        msg = get_agent_executor(the_anthropic_model=True).invoke(
            {"messages": [the_message]}
        )





        the_last_messages = msg["messages"]









        return_value = the_last_messages[-1].content
        if isinstance(return_value, list):
            the_text = ""
            for each in return_value:
                the_text += str(each)
            return_value = the_text

        if return_value == "":
            return_value = "No response "





        return return_value

    except Exception as e:
        traceback.print_exc()




try:
    from ..llm import get_model
    from ..utils.db import *
    from ..llm_settings import llm_settings
    from ..tooler import *
    from ..display_tools import *
    from ..cu.computer import *
    from ..teams import *
    from .agent_tools import get_tools
    from ..mcp.tool import mcp_tools
    from ..standard_tools import get_standard_tools
    
except ImportError:
    from llm import get_model
    from utils.db import *
    from llm_settings import llm_settings
    from tooler import *
    from display_tools import *
    from cu.computer import *
    from teams import *
    from agent.agent_tools import get_tools
    from mcp.tool import mcp_tools
    from standard_tools import get_standard_tools


from langgraph.prebuilt import chat_agent_executor


custom_tools_ = []


def custom_tools():
    global custom_tools_
    the_list = []
    the_list += custom_tools_
    return the_list


prompt_cache = {}


def get_prompt(name):
    global prompt_cache
    if name in prompt_cache:
        return prompt_cache[name]
    else:
        from langchain import hub

        prompt = hub.pull(name)
        prompt_cache[name] = prompt
        return prompt


def get_agent_executor(the_anthropic_model=False, no_tools=False):
    tools = get_tools()
    tools += custom_tools()

    model = load_model_settings()

    if is_predefined_agents_setting_active() and llm_settings[model]["tools"]:
        try:
            import crewai

            tools += [search_on_internet_and_report_team, generate_code_with_aim_team]
        except ImportError:
            pass


    if the_anthropic_model:
        tools += []
        if load_aws_access_key_id() == "default":
            model_catch = get_model(the_model="claude-3-5-sonnet-20241022")
        else:
            model_catch = get_model(the_model="us.anthropic.claude-3-5-sonnet-20241022-v2:0")

        print("Anthropic model catch", model_catch)
        print("Anthropic tools len", len(tools))
        return chat_agent_executor.create_tool_calling_executor(model_catch, tools)
    else:
        tools += [mouse_scroll, click_to_text, click_to_icon, click_to_area, screenshot] + mcp_tools() + get_standard_tools()



    if no_tools:
        tools = []


    if (
        llm_settings[model]["provider"] == "openai"
        or llm_settings[model]["provider"] == "groq"
        or llm_settings[model]["provider"] == "azureai"
        or llm_settings[model]["provider"] == "anthropic"
        or llm_settings[model]["provider"] == "aws"


    ):
        return chat_agent_executor.create_tool_calling_executor(get_model(), tools)

    if llm_settings[model]["provider"] == "ollama":
        print("Ollama tool len", len(tools))
        return chat_agent_executor.create_tool_calling_executor(get_model(), tools)



import soundfile as sf


synthesiser_ = None


def synthesiser():
    from transformers import pipeline

    global synthesiser_
    if synthesiser_ is None:
        synthesiser_ = pipeline("text-to-speech", "microsoft/speecht5_tts")
    return synthesiser_


embeddings_dataset_ = None


def embeddings_dataset():
    from datasets import load_dataset

    global embeddings_dataset_
    if embeddings_dataset_ is None:
        embeddings_dataset_ = load_dataset(
            "Matthijs/cmu-arctic-xvectors", split="validation"
        )
    return embeddings_dataset_


speaker_embedding_ = None


def speaker_embedding():
    import torch

    global speaker_embedding_
    if speaker_embedding_ is None:
        speaker_embedding_ = torch.tensor(
            embeddings_dataset()[7306]["xvector"]
        ).unsqueeze(0)
    return speaker_embedding_


def preload_tts_microsoft_local():
    synthesiser()
    embeddings_dataset()
    speaker_embedding()


def tts_microsoft_local(text_chunk, location):
    speech = synthesiser()(
        text_chunk, forward_params={"speaker_embeddings": speaker_embedding()}
    )
    sf.write(location, speech["audio"], samplerate=speech["sampling_rate"])
    return location




from kot import KOT

try:
    from .folder import currently_dir, artifacts_dir, media_dir
except:
    from folder import currently_dir, artifacts_dir, media_dir









from langchain_core.messages import HumanMessage, SystemMessage, AIMessage


try:
    from .db import *

except:
    from db import *


import time



class Human:
    def __init__(self, content, the_time, auto_delete:int=None):

        self.that_was_empty = False

        if isinstance(content, dict):
            if "text" in content:
                if content["text"] == "":
                    self.that_was_empty = True
                    content["text"] = "No response"

        if isinstance(content, list):
            for i in range(len(content)):
                if "text" in content[i]:
                    if content[i]["text"] == "":
                        self.that_was_empty = True
                        content[i]["text"] = "No response"


        self.content = content
        self.timestamp = the_time
        self.auto_delete = auto_delete

    def __dict__(self):
        current_time = time.time()

        if self.auto_delete is not None:
            print(current_time, self.timestamp, self.auto_delete)
            if current_time - self.timestamp > self.auto_delete:
                return {"type": "human", "content": "This content deleted.", "timestamp": self.timestamp, "auto_delete": self.auto_delete, "that_was_empty": self.that_was_empty}

        return {"type": "human", "content": self.content, "timestamp": self.timestamp, "auto_delete": self.auto_delete, "that_was_empty": self.that_was_empty}

class Assistant:
    def __init__(self, content, the_time):

        self.that_was_empty = False

        if isinstance(content, dict):
            if "text" in content:
                if content["text"] == "":
                    self.that_was_empty = True
                    content["text"] = "No response"

        if isinstance(content, list):
            for i in range(len(content)):
                if "text" in content[i]:
                    if content[i]["text"] == "":
                        self.that_was_empty = True
                        content[i]["text"] = "No response"
        self.content = content
        self.timestamp = the_time

    def __dict__(self):
        return {"type": "assistant", "content": self.content, "timestamp": self.timestamp, "that_was_empty": self.that_was_empty}

class System:
    def __init__(self, content, the_time):

        self.that_was_empty = False

        if isinstance(content, dict):
            if "text" in content:
                if content["text"] == "":
                    self.that_was_empty = True
                    content["text"] = "No response"

        if isinstance(content, list):
            for i in range(len(content)):
                if "text" in content[i]:
                    if content[i]["text"] == "":
                        self.that_was_empty = True
                        content[i]["text"] = "No response"


        self.content = content
        self.timestamp = the_time

    def __dict__(self):
        return {"type": "system", "content": self.content, "timestamp": self.timestamp, "that_was_empty": self.that_was_empty}


class ChatHistory:

    def __init__(self):
        self.chat_id = get_profile()
        self.db = KOT(f"chat_history_{self.chat_id}", folder=artifacts_dir, enable_hashing=True)

        if self.db.get("chat") is None:
            self.db.set("chat", [])
  
        if self.get_chat() == []:
            print("SETTING CHAT")
            self.add_message("system", {"type":"text", "text": load_system_prompt()})


    def add_message(self, message_type:str, content, auto_delete:int=None):

        the_time = time.time()


        if content == []:
            content = {"type":"text", "text": "No response"}


        if message_type == "human":
            message = Human(content, the_time, auto_delete)
        elif message_type == "assistant":
            print("ASSISTANT", content)
            message = Assistant(content, the_time)
        elif message_type == "system":
            print("SYSTEM", content)
            message = System(content, the_time)
        else:
            raise ValueError("Invalid message type")

       
        chat = self.db.get("chat")
        chat.append(message.__dict__())


        self.db.set("chat", chat)

    def get_chat(self):
        chat = self.db.get("chat")
        chat = sorted(chat, key=lambda x: x["timestamp"])

        print("CHAT", chat)
        # Transform dict to Message objects

        the_chat = []
        for message in chat:
            if message["type"] == "human":
                the_chat.append(Human(content=message["content"], the_time=message["timestamp"], auto_delete=message["auto_delete"]))
            elif message["type"] == "assistant":
                the_chat.append(Assistant(content=message["content"], the_time=message["timestamp"]))
            elif message["type"] == "system":
                the_chat.append(System(content=message["content"], the_time=message["timestamp"]))

        
        last_chat = []
        for message in the_chat:
            if message.that_was_empty:
                continue
            last_chat.append(message.__dict__())


        chat = last_chat

        langchain_messages = []

        for message in chat:

            if isinstance(message["content"], tuple):
                message["content"] = list(message["content"])
            if isinstance(message["content"], dict):
                message["content"] = [message["content"]]
            








            if message["type"] == "human":
                
                langchain_messages.append(HumanMessage(content=
                    message["content"]
                    ))
            elif message["type"] == "assistant":
                langchain_messages.append(AIMessage(content=
                    message["content"]
                    ))
            elif message["type"] == "system":
                langchain_messages.append(SystemMessage(content=
                    message["content"]
                    ))


        return langchain_messages
    

    def clear_chat(self):
        self.db.set("chat", [])

    

        



from PyQt5.QtWidgets import QDialog, QVBoxLayout, QLabel, QLineEdit, QPushButton
from PyQt5.QtCore import Qt

try:
    from ..utils.db import *
    from ..agent.chat_history import clear_chat_history
except:
    from utils.db import *
    from agent.chat_history import clear_chat_history


def settings_popup(self):
    """
    Display a settings popup dialog for configuring various options.

    This function creates a settings dialog with options to reset chat history, enable/disable the just text model,
    and change the active profile.

    Parameters:
    - self: Reference to the main application window.

    Returns:
    - None
    """
    from ..gpt_computer_assistant import the_main_window

    settings_dialog = QDialog()
    settings_dialog.setWindowTitle("Settings")
    settings_dialog.setWindowModality(Qt.ApplicationModal)

    settings_dialog.setLayout(QVBoxLayout())

    reset_memory_button = QPushButton("Reset Memory")

    def clear_chat_history_():
        """
        Clear the chat history and update the main window.

        This function clears the chat history and updates the main window with a notification.

        Returns:
        - None
        """
        clear_chat_history()
        the_main_window.update_from_thread("Cleared Chat History")
        settings_dialog.close()

    reset_memory_button.clicked.connect(clear_chat_history_)
    settings_dialog.layout().addWidget(reset_memory_button)

    just_text_button = QPushButton("Enable Just Text Model")

    settings_dialog.layout().addWidget(just_text_button)

    if is_just_text_model_active():
        just_text_button.setText("Disable Just Text Model")

        def deactivate_just_text_model_():
            """
            Deactivate the just text model and update the main window.

            This function deactivates the just text model and updates the main window with a notification.

            Returns:
            - None
            """
            deactivate_just_text_model()
            the_main_window.update_from_thread("Disabled Just Text Model")
            settings_dialog.close()

        just_text_button.clicked.connect(deactivate_just_text_model_)
    else:

        def activate_just_text_model_():
            """
            Activate the just text model and update the main window.

            This function activates the just text model and updates the main window with a notification.

            Returns:
            - None
            """
            activate_just_text_model()
            the_main_window.update_from_thread("Enabled Just Text Model")
            settings_dialog.close()

        just_text_button.clicked.connect(activate_just_text_model_)

    settings_dialog.layout().addWidget(QLabel("Profile"))
    profile_input = QLineEdit()

    profile_input.setText(get_profile())
    settings_dialog.layout().addWidget(profile_input)
    profile_save_button = QPushButton("Save")

    def set_profile_(profile):
        """
        Set the active profile and update the main window.

        This function sets the active profile based on user input and updates the main window with a notification.

        Parameters:
        - profile (str): The profile name to set.

        Returns:
        - None
        """
        set_profile(profile)
        the_main_window.update_from_thread("Saved Profile")
        settings_dialog.close()

    profile_save_button.clicked.connect(lambda: set_profile_(profile_input.text()))
    settings_dialog.layout().addWidget(profile_save_button)

    dark_mode_button = QPushButton("Enable Dark Mode")

    settings_dialog.layout().addWidget(dark_mode_button)

    if is_dark_mode_active():
        dark_mode_button.setText("Disable Dark Mode")

        def deactivate_dark_mode_():
            """
            Deactivate dark mode and update the main window.

            This function deactivates dark mode and updates the main window with a notification.

            Returns:
            - None
            """
            deactivate_dark_mode()
            the_main_window.update_from_thread("Disabled Dark Mode")
            the_main_window.light_mode()
            settings_dialog.close()

        dark_mode_button.clicked.connect(deactivate_dark_mode_)
    else:

        def activate_dark_mode_():
            """
            Activate dark mode and update the main window.

            This function activates dark mode and updates the main window with a notification.

            Returns:
            - None
            """
            activate_dark_mode()
            the_main_window.update_from_thread("Enabled Dark Mode")
            the_main_window.dark_mode()
            settings_dialog.close()

        dark_mode_button.clicked.connect(activate_dark_mode_)

    predefined_agents_button = QPushButton(
        "Enable Predefined Agents (Good Results, Long Response Time)"
    )

    settings_dialog.layout().addWidget(predefined_agents_button)

    try:
        if is_predefined_agents_setting_active():
            predefined_agents_button.setText(
                "Disable Predefined Agents (Bad Results, Short Response Time)"
            )

            def deactivate_predefined_agents_():
                deactivate_predefined_agents_setting()
                the_main_window.update_from_thread(
                    "Disabled Predefined Agents (Bad Results, Short Response Time)"
                )
                settings_dialog.close()

            predefined_agents_button.clicked.connect(deactivate_predefined_agents_)
        else:

            def activate_predefined_agents_():
                activate_predefined_agents_setting()
                the_main_window.update_from_thread(
                    "Enabled Predefined Agents (Good Results, Long Response Time)"
                )
                settings_dialog.close()

            predefined_agents_button.clicked.connect(activate_predefined_agents_)

    except:
        predefined_agents_button.setText("Install gpt-computer-assistant[agentic]")

    online_tools_button = QPushButton(
        "Enable Upsonic Tiger Tools - More Capability (Recommended)"
    )

    settings_dialog.layout().addWidget(online_tools_button)

    if is_online_tools_setting_active():
        online_tools_button.setText(
            "Disable Upsonic Tiger Tools - Low Capability (Not Recommended)"
        )

        def deactivate_online_tools_():
            deactivate_online_tools_setting()
            the_main_window.update_from_thread(
                "Disabled Upsonic Tiger Tools - Low Capability (Not Recommended)"
            )
            settings_dialog.close()

        online_tools_button.clicked.connect(deactivate_online_tools_)
    else:

        def activate_online_tools_():
            activate_online_tools_setting()
            the_main_window.update_from_thread(
                "Enabled Upsonic Tiger Tools - More Capability (Recommended)"
            )
            settings_dialog.close()

        online_tools_button.clicked.connect(activate_online_tools_)

    auto_stop_recording_button = QPushButton("Enable Auto Stop Recording")

    settings_dialog.layout().addWidget(auto_stop_recording_button)

    if is_auto_stop_recording_setting_active():
        auto_stop_recording_button.setText("Disable Auto Stop Recording")

        def deactivate_auto_stop_recording_():
            deactivate_auto_stop_recording_setting()
            the_main_window.update_from_thread("Disabled Auto Stop Recording")
            settings_dialog.close()

        auto_stop_recording_button.clicked.connect(deactivate_auto_stop_recording_)
    else:

        def activate_auto_stop_recording_():
            activate_auto_stop_recording_setting()
            the_main_window.update_from_thread("Enabled Auto Stop Recording")
            settings_dialog.close()

        auto_stop_recording_button.clicked.connect(activate_auto_stop_recording_)

    api_key_label = QLabel("Wakeword - Pvporcupine API Key")
    settings_dialog.layout().addWidget(api_key_label)
    api_key_input = QLineEdit()
    api_key = load_pvporcupine_api_key()
    api_key_input.setText(api_key)
    settings_dialog.layout().addWidget(api_key_input)
    save_button = QPushButton("Save")

    def save_api_key_(api_key):
        first_time = True
        if api_key != "CHANGE_ME":
            first_time = False
        save_pvporcupine_api_key(api_key)

        the_main_window.update_from_thread(
            "Wake word activated, just say 'Her Computer' or jarvis to activate the assistant"
        )
        if first_time:
            the_main_window.wake_word_trigger()
        settings_dialog.close()

    save_button.clicked.connect(lambda: save_api_key_(api_key_input.text()))
    settings_dialog.layout().addWidget(save_button)

    wake_word_button = QPushButton("Enable Wake Word")

    settings_dialog.layout().addWidget(wake_word_button)

    missing_parts = False
    try:
        pass
    except:
        missing_parts = True

    if api_key == "CHANGE_ME":
        wake_word_button.setText("Please Set Pvporcupine API Key First")
    elif missing_parts:
        wake_word_button.setText("Please Install gpt-computer-assistant[wakeword]")
    else:
        if is_wake_word_active():
            wake_word_button.setText("Disable Wake Word")

            def deactivate_wake_word_():
                deactivate_wake_word()
                the_main_window.update_from_thread("Disabled Wake Word")
                the_main_window.wake_word_active = False
                settings_dialog.close()

            wake_word_button.clicked.connect(deactivate_wake_word_)
        else:

            def activate_wake_word_():
                activate_wake_word()
                the_main_window.update_from_thread("Enabled Wake Word")
                the_main_window.wake_word_active = True
                the_main_window.wake_word_trigger()
                settings_dialog.close()

            wake_word_button.clicked.connect(activate_wake_word_)

    wake_word_screen_button = QPushButton("Enable Screen Input for Wake Word Mode")

    settings_dialog.layout().addWidget(wake_word_screen_button)

    if is_wake_word_screen_setting_active():
        wake_word_screen_button.setText("Disable Screen Input for Wake Word Mode")

        def deactivate_auto_stop_recording_():
            deactivate_wake_word_screen_setting()
            the_main_window.update_from_thread(
                "Disabled Screen Input for Wake Word Mode"
            )
            settings_dialog.close()

        wake_word_screen_button.clicked.connect(deactivate_auto_stop_recording_)
    else:

        def activate_auto_stop_recording_():
            activate_wake_word_screen_setting()
            the_main_window.update_from_thread(
                "Enabled Screen Input for Wake Word Mode"
            )
            settings_dialog.close()

        wake_word_screen_button.clicked.connect(activate_auto_stop_recording_)

    continuously_conversations_button = QPushButton("Enable Continuously Conversations")

    settings_dialog.layout().addWidget(continuously_conversations_button)

    if is_continuously_conversations_setting_active():
        continuously_conversations_button.setText("Disable Continuously Conversations")

        def deactivate_auto_stop_recording_():
            deactivate_continuously_conversations_setting()
            the_main_window.update_from_thread("Disabled Continuously Conversations")
            settings_dialog.close()

        continuously_conversations_button.clicked.connect(
            deactivate_auto_stop_recording_
        )
    else:

        def activate_auto_stop_recording_():
            activate_continuously_conversations_setting()
            the_main_window.update_from_thread("Enabled Continuously Conversations")
            settings_dialog.close()

        continuously_conversations_button.clicked.connect(activate_auto_stop_recording_)

    settings_dialog.exec_()



def string_to_function(func_str):
    # Define a local dictionary to execute the function string
    local_dict = {}

    # Execute the function string in the local dictionary
    exec(func_str, globals(), local_dict)

    # Extract the function name
    func_name = func_str.split("(")[0].split()[-1]

    # Return the function object from the local dictionary
    return local_dict[func_name]



from langchain_community.chat_message_histories import SQLChatMessageHistory
from .background import llm_history_oiginal

try:
    from ..utils.db import get_history_db
    from ..utils.db import load_model_settings, agents
    from ..llm_settings import llm_settings
    from ..utils.chat_history import ChatHistory
except ImportError:
    from utils.db import get_history_db
    from utils.db import load_model_settings
    from utils.chat_history import ChatHistory
    from llm_settings import llm_settings


def get_chat_message_history():
    connection = SQLChatMessageHistory(
        session_id="abc123", connection_string=f"sqlite:///{get_history_db()}"
    )
    if len(connection.messages) == 0:
        the_model = load_model_settings()
        if llm_settings[the_model]["tools"]:
            connection.add_message(llm_history_oiginal()[0])

    return connection


def clear_chat_history():
    ChatHistory().clear_chat()



try:
    from ..utils.db import *
    from ..tooler import *
    from ..display_tools import *
    from ..teams import *
    from ..llm_settings import llm_settings
except ImportError:
    from utils.db import *

    from tooler import *
    from display_tools import *
    from teams import *
    from llm_settings import llm_settings


custom_tools = []


def load_tiger_tools():
    try:
        from upsonic import Tiger

        tools = Tiger()
        tools.enable_auto_requirements = True
        tools = tools.langchain()
        return tools
    except:
        return False


def load_default_tools():
    from ..standard_tools import get_standard_tools

    return get_standard_tools()


cached_tiger_tools = None


def get_tiger_tools():
    global cached_tiger_tools
    if cached_tiger_tools is None:
        cached_tiger_tools = load_tiger_tools()
    return cached_tiger_tools


if is_online_tools_setting_active():
    get_tiger_tools()


def get_tools():
    return []



"""Callback Handler streams to stdout on new llm token."""

from langchain.callbacks.streaming_stdout_final_only import (
    FinalStreamingStdOutCallbackHandler,
)
from typing import Any


class customcallback(FinalStreamingStdOutCallbackHandler):
    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        self.append_to_last_tokens(token)

        if self.check_if_answer_reached():
            self.answer_reached = True

            return

        if self.answer_reached:
            from .gpt_computer_assistant import the_main_window

            the_main_window.set_text_to_input_box(token)



try:
    from ..utils.db import *
    from ..llm import get_model
    from ..top_bar_wrapper import wrapper
    from ..llm_settings import llm_settings
except ImportError:
    from utils.db import *
    from top_bar_wrapper import wrapper
    from llm_settings import llm_settings

from langchain.tools import tool


import base64
import math
import os
import platform
import shlex
import shutil
import tempfile
import time
from strenum import StrEnum
from pathlib import Path
from typing import Literal, TypedDict
from uuid import uuid4
import pyautogui
from anthropic.types.beta import BetaToolComputerUse20241022Param
from PIL import Image

from .base import BaseAnthropicTool, ToolError, ToolResult
from .run import run

OUTPUT_DIR = "/tmp/outputs"

TYPING_DELAY_MS = 12
TYPING_GROUP_SIZE = 50

Action = Literal[
    "key",
    "type",
    "mouse_move",
    "left_click",
    "left_click_drag",
    "right_click",
    "middle_click",
    "double_click",
    "screenshot",
    "cursor_position",
]


class Resolution(TypedDict):
    width: int
    height: int


MAX_SCALING_TARGETS: dict[str, Resolution] = {
    "XGA": Resolution(width=1024, height=768),
    "WXGA": Resolution(width=1280, height=800),
    "FWXGA": Resolution(width=1366, height=768),
}


class ScalingSource(StrEnum):
    COMPUTER = "computer"
    API = "api"


class ComputerToolOptions(TypedDict):
    display_height_px: int
    display_width_px: int
    display_number: int | None


def chunks(s: str, chunk_size: int) -> list[str]:
    return [s[i : i + chunk_size] for i in range(0, len(s), chunk_size)]


def smooth_move_to(x, y, duration=1.2):

    pyautogui.moveTo(x, y)


def key_action(text: str):
    if platform.system() == "Darwin":
        text = text.replace("super+", "command+")

    def normalize_key(key):
        key = key.lower().replace("_", "")
        key_map = {
            "pagedown": "pgdn",
            "pageup": "pgup",
            "enter": "return",
            "return": "enter",
        }
        return key_map.get(key, key)

    keys = [normalize_key(k) for k in text.split("+")]

    if len(keys) > 1:
        if "darwin" in platform.system().lower():
            keystroke, modifier = (keys[-1], "+".join(keys[:-1]))
            modifier = modifier.lower() + " down"
            if keystroke.lower() == "space":
                keystroke = " "
            elif keystroke.lower() == "enter":
                keystroke = "\n"
            script = f"""
            tell application "System Events"
                keystroke "{keystroke}" using {modifier}
            end tell
            """
            os.system("osascript -e '{}'".format(script))
        else:
            pyautogui.hotkey(*keys)
    else:
        pyautogui.press(keys[0])


def type_action(text: str):
    pyautogui.write(text, interval=TYPING_DELAY_MS / 1000)


def click_action(action: str):
    time.sleep(0.1)
    button = {
        "left_click": "left",
        "right_click": "right",
        "middle_click": "middle",
    }
    if action == "double_click":
        pyautogui.click()
        time.sleep(0.1)
        pyautogui.click()
    else:
        pyautogui.click(button=button.get(action, "left"))


def screenshot_action(direct_base64: bool = False) -> ToolResult:
    """
    See the screenshot of the current screen.
    """


    temp_dir = Path(tempfile.gettempdir())
    path = temp_dir / f"screenshot_{uuid4().hex}.png"

    screenshot = pyautogui.screenshot()
    screenshot.save(str(path))

    if _scaling_enabled:
        x, y = scale_coordinates(ScalingSource.COMPUTER, width, height)
        print(f"Scaling screenshot to {x}x{y}")
        with Image.open(path) as img:
            img = img.resize((x, y), Image.Resampling.LANCZOS)
            img.save(path)

    if path.exists():

        with open(path, "rb") as image_file:
            base64_image = base64.b64encode(image_file.read()).decode("utf-8")

        path.unlink()
        if direct_base64:
            return base64_image
        return ToolResult(base64_image=base64_image)
    raise ToolError(f"Failed to take screenshot")


def screenshot_action_(path):


    screenshot = pyautogui.screenshot()
    screenshot.save(str(path))

    if _scaling_enabled:
        x, y = scale_coordinates(ScalingSource.COMPUTER, width, height)
        print(f"Scaling screenshot to {x}x{y}")
        with Image.open(path) as img:
            img = img.resize((x, y), Image.Resampling.LANCZOS)
            img.save(path)




def cursor_position_action():
    """
    Get the current position of the cursor as (x, y).
    """

    x, y = pyautogui.position()
    x, y = scale_coordinates(ScalingSource.COMPUTER, x, y)
    return ToolResult(output=f"X={x},Y={y}")


def shell_action(command: str, take_screenshot=True) -> ToolResult:
    """
    Run a shell command and return the output.
    """
    _, stdout, stderr = run(command)
    base64_image = None

    if take_screenshot:
        time.sleep(_screenshot_delay)
        base64_image = screenshot_action().base64_image

    return ToolResult(output=stdout, error=stderr, base64_image=base64_image)


def scale_coordinates(source: ScalingSource, x: int, y: int):
    if not _scaling_enabled:
        return x, y
    ratio = width / height
    target_dimension = None
    for dimension in MAX_SCALING_TARGETS.values():
        if abs(dimension["width"] / dimension["height"] - ratio) < 0.02:
            if dimension["width"] < width:
                target_dimension = dimension
            break
    if target_dimension is None:
        return x, y
    x_scaling_factor = target_dimension["width"] / width
    y_scaling_factor = target_dimension["height"] / height
    if source == ScalingSource.API:
        if x > width or y > height:
            raise ToolError(f"Coordinates {x}, {y} are out of bounds")
        return round(x / x_scaling_factor), round(y / y_scaling_factor)
    return round(x * x_scaling_factor), round(y * y_scaling_factor)


@wrapper
def mouse_move_action(coordinate: tuple[int, int]):
    """Move the mouse to the specified coordinate."""
    if coordinate is None:
        raise ToolError("coordinate is required for mouse_move")
    x, y = scale_coordinates(ScalingSource.API, coordinate[0], coordinate[1])
    smooth_move_to(x, y)

@wrapper
def left_click_drag_action(coordinate: tuple[int, int]):
    """Perform a left click and drag to the specified coordinate."""
    if coordinate is None:
        raise ToolError("coordinate is required for left_click_drag")
    x, y = scale_coordinates(ScalingSource.API, coordinate[0], coordinate[1])
    smooth_move_to(x, y)
    pyautogui.dragTo(x, y, button="left")

@wrapper
def key_action_handler(text: str):
    """Press a specific key."""
    if text is None:
        raise ToolError("text is required for key")
    key_action(text)

@wrapper
def type_action_handler(text: str):
    """Type the specified text."""
    if text is None:
        raise ToolError("text is required for type")
    type_action(text)

@wrapper
def left_click_action():
    """Perform a left click."""
    click_action("left_click")

@wrapper
def right_click_action():
    """Perform a right click."""
    click_action("right_click")

@wrapper
def middle_click_action():
    """Perform a middle click."""
    click_action("middle_click")

@wrapper
def double_click_action():
    """Perform a double click."""
    click_action("double_click")








# Initialize global variables
width, height = pyautogui.size()

display_num = None
_screenshot_delay = 2.0
_scaling_enabled = True



computer_tool = [tool(mouse_move_action), tool(left_click_drag_action), tool(key_action_handler), tool(type_action_handler), tool(left_click_action), tool(right_click_action), tool(middle_click_action), tool(double_click_action), tool(screenshot_action), tool(cursor_position_action), tool(shell_action)]



import base64
import pyautogui

try:
    from ..gui.signal import signal_handler
    from ..utils.db import just_screenshot_path
    from ..cu.computer import screenshot_action_
except ImportError:
    from gui.signal import signal_handler
    from utils.db import just_screenshot_path
    from cu.computer import screenshot_action_


def encode_image(image_path):
    """
    Encode an image file to base64 format.

    Parameters:
    - image_path (str): The path to the image file to encode.

    Returns:
    - str or None: The base64 encoded string of the image, or None if an error occurs.
    """
    try:
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode("utf-8")
    except FileNotFoundError:
        print(f"File not found: {image_path}")
        return None
    except Exception as e:
        print(f"An error occurred while encoding the image: {e}")
        return None


def take_screenshot():
    """
    Take a screenshot using pyautogui and save it.

    This function takes a screenshot of the entire screen using pyautogui,
    saves it to the specified path, and emits a signal indicating that
    the assistant is thinking.

    Returns:
    - None
    """
    try:
        screenshot_action_(just_screenshot_path)
        signal_handler.assistant_thinking.emit()
    except Exception as e:
        print(f"An error occurred while taking the screenshot: {e}")



import traceback
from .remote import Remote_Client
import json
import re
import hashlib


def extract_json(llm_output):

    # Use regex to extract the json then transform it to a python dict object ```json ````
    json_str = re.search(r'```json\n(.*?)```', llm_output, re.DOTALL).group(1)


    # transform the json string to a python dict object
    transformed_json = json.loads(json_str)



    return transformed_json



    
    

class BaseClass:
    def __init__(self, screen_task=False):
        self.screen_task = screen_task

    def add_client(self, client: Remote_Client):

        self.client = client

        if hasattr(self, "verifier"):

            self.verifier.add_client(client)

    def add_task(self, task):
        self.task = task




    def sha_hash(self, text):
        
        return hashlib.sha256(text.encode()).hexdigest()



class BaseVerifier(BaseClass):
    def __init__(self, try_count=5, exception_return=None, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.try_count = try_count
        self.exception_return = exception_return


class TypeVerifier(BaseVerifier):
    def __init__(self, type, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.type = type

    def verify(self, description, result):



        control_point = self.client.request(
            f"""
User Request:
{description}

AI Result:
{result}


Now critically analyze the result of the task you just completed. If the request is impossible to complete you can respond with "This task is impossible to complete" and "Reason:" to stop the task.

Getting current state:
- See the screen
- Read the history of conversation to understand the context of the task.


If the result fails respond onyl with “I am sorry” and "Reason:" to trigger a retry.
            """
            , "", screen=self.screen_task)


        if "I am sorry" in control_point:

            raise Exception(f"Not satisfied with the result {self.task.description}")


        
        prompt = """
        Hi, now your responsibility is returning the answer in the requested format.

        User only wants the result in the format of """f"""{self.type}"""+""".
        Dont use any other format or any other type of data.

        Format recipe:
        1. list
            Return the user want like this:
            ```json
            ["element1", "element2", "element3"]
            ```

        2. dict
            Return the user want like this:
            ```json
            {"key1": "value1", "key2": "value2", "key3": "value3"}
            ```
        
        3. list in list
            Return the user want like this:
            ```json
            [["element1", "element2", "element3"], ["element4", "element5", "element6"]]
            ```

        4. dict in list
            Return the user want like this:
            ```json
            [{"key1": "value1", "key2": "value2", "key3": "value3"}, {"key4": "value4", "key5": "value5", "key6": "value6"}]
            ```
        
        5. string
            Return the user want like this:
            ```json
            "This is a string"
            ```

        6. integer
            Return the user want like this:
            ```json
            123
            ```
        
        7. float
            Return the user want like this:
            ```json
            123.456
            ```
        
        8. bool
            Return the user want like this:
            ```json
            true
            ```

            

        End of the day return result in ```json ``` format.
        """

        self.client.change_profile(self.task.hash)
        result = self.client.request(prompt, "", screen=self.screen_task)





        result = extract_json(result)
        return result



            


class Task(BaseClass):
    def __init__(self, description, verifier: BaseVerifier = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.description = description
        self.output = ""
        self.verifier = verifier
        if self.verifier:
            self.verifier.add_task(self)

        self.result = None

        self.hash = self.sha_hash(description)



    def run(self):


        

        # Verify the result
        if self.verifier:

            try_count = 0

            while try_count < self.verifier.try_count:
                try_count += 1
                if try_count > 1:
                    self.output = "User is not satisfied with the result. Please try again."
                self.client.change_profile(self.hash)
                result = self.client.request(self.description, self.output, screen=self.screen_task)

                try:
                    self.client.change_profile(self.hash+"VERIFY")
                    result = self.verifier.verify(self.description, result)
                    break
                except:
                    traceback.print_exc()

                    result = self.verifier.exception_return
                    

            



        self.result = result
        return result



import random


from langchain_core.messages import HumanMessage, SystemMessage, AIMessage

from .chat_history import *
from .agent import *


try:
    from ..screen.shot import *
    from ..utils.db import load_model_settings, agents
    from ..llm import get_model
    from ..llm_settings import llm_settings
    from ..utils.chat_history import ChatHistory
except ImportError:
    from screen.shot import *
    from utils.db import load_model_settings, agents
    from utils.chat_history import ChatHistory
    from llm import get_model
    from llm_settings import llm_settings

config = {"configurable": {"thread_id": "abc123"}, "recursion_limit":10}



def random_charachter(length=10):
    return "".join(random.choices("abcdefghijklmnopqrstuvwxyz0123456789", k=length))



def agentic(
    llm_input, llm_history, client, screenshot_path=None, dont_save_image=False
):
    global agents
    from crewai import Task, Crew

    from crewai import Agent as crewai_Agent

    the_agents = []

    for each in agents:
        the_agents.append(
            crewai_Agent(
                role=each["role"],
                goal=each["goal"],
                backstory=each["backstory"],
                llm=get_model(high_context=True),
            )
        )

    agents = the_agents

    print("LLM INPUT", llm_input)

    def image_explaination():
        the_message = [
            {"type": "text", "text": "Explain the image"},
        ]

        if screenshot_path:
            base64_image = encode_image(screenshot_path)
            the_message.append(
                {
                    "type": "image_url",
                    "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"},
                },
            )
            print("LEN OF İMAGE", len(base64_image))

        the_message = HumanMessage(content=the_message)
        get_chat_message_history().add_message(the_message)

        the_model = load_model_settings()

        if (
            llm_settings[the_model]["provider"] == "openai"
            or llm_settings[the_model]["provider"] == "ollama"
            or llm_settings[the_model]["provider"] == "azureai"
            or llm_settings[the_model]["provider"] == "anthropic"
            or llm_settings[the_model]["provider"] == "aws"
        ):
            msg = get_agent_executor().invoke(
                {"messages": llm_history + [the_message]}, config=config
            )

        if llm_settings[the_model]["provider"] == "google":
            msg = get_agent_executor().invoke(
                {"messages": llm_history + [the_message]}, config=config
            )

        the_last_messages = msg["messages"]

        return the_last_messages[-1].content

    if screenshot_path:
        image_explain = image_explaination()
        llm_input += "User Sent Image and image content is: " + image_explain

    llm_input = llm_input 

    task = Task(
        description=llm_input,
        expected_output="Answer",
        agent=agents[0],
        tools=get_tools(),
    )

    the_crew = Crew(
        agents=agents,
        tasks=[task],
        full_output=True,
        verbose=True,
    )

    result = the_crew.kickoff()["final_output"]

    get_chat_message_history().add_message(
        HumanMessage(content=[llm_input])
    )
    get_chat_message_history().add_message(AIMessage(content=[result]))

    return result


def assistant(
    llm_input, client, screenshot_path=None, dont_save_image=False, just_screenshot=False
):
    
    the_chat_history = ChatHistory()

    the_model = load_model_settings()

    print("LLM INPUT", llm_input)

    if llm_settings[the_model]["tools"]:
        llm_input = llm_input 


    human_first_message = {"type": "text", "text": f"{llm_input}"}
    the_chat_history.add_message("human", human_first_message)


    the_message = [
        human_first_message
    ]



    human_second_message = None

    if screenshot_path:
        base64_image = encode_image(screenshot_path)
        if llm_settings[the_model]["provider"] == "ollama":

            human_second_message = {
                "type": "image_url",
                "image_url": base64_image,
            }
            the_chat_history.add_message("human", human_second_message, auto_delete=10)


        else:
            human_second_message = {
                "type": "image_url",
                "image_url": {"url": f"data:image/png;base64,{base64_image}"},
            }
            the_chat_history.add_message("human", human_second_message, auto_delete=10)


        print("LEN OF IMAGE", len(base64_image))


    if human_second_message:
        the_message.append(human_second_message)





    the_message = HumanMessage(content=the_message)


    

    llm_history = the_chat_history.get_chat()




    

    if (
        llm_settings[the_model]["provider"] == "openai"
        or llm_settings[the_model]["provider"] == "azureai"
        or llm_settings[the_model]["provider"] == "anthropic"
        or llm_settings[the_model]["provider"] == "aws"
    ):
        if just_screenshot:
            msg = {"messages": llm_history + [the_message]}
            time.sleep(1)
        else:
            print("The model:", the_model)
            msg = get_agent_executor().invoke(
                {"messages": llm_history + [the_message]}, config=config
            )
      




    the_last_messages = msg["messages"]



    the_chat_history.add_message("assistant", the_last_messages[-1].content)






    return_value = the_last_messages[-1].content
    if isinstance(return_value, list):
        the_text = ""
        for each in return_value:
            the_text += str(each)
        return_value = the_text

    if return_value == "":
        return_value = "No response "
        return_value += str(random_charachter())


    if just_screenshot:
        return "OK"

    return return_value



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from setuptools import setup
import platform


# Read the requirements from the requirements.txt file
with open("requirements.txt") as fp:
    install_requires = fp.read().splitlines()

if platform.system() in ["Windows"]:
    install_requires.append("AppOpener==1.7")

elif platform.system() == "Darwin":  # Darwin is the system name for macOS
    install_requires.append(
        "MacAppOpener==0.0.5"
    )  # Replace with actual macOS specific package


setup(
    name="gpt_computer_assistant",
    version="0.26.11",
    description="""GPT""",
    long_description="".join(open("README.md", encoding="utf-8").readlines()),
    long_description_content_type="text/markdown",
    url="https://github.com/onuratakan/gpt-computer-assistant",
    author="Upsonic",
    author_email="onur@upsonic.co",
    license="MIT",
    packages=[
        "gpt_computer_assistant",
        "gpt_computer_assistant.agent",
        "gpt_computer_assistant.cu",
        "gpt_computer_assistant.mcp",
        "gpt_computer_assistant.gui",
        "gpt_computer_assistant.screen",
        "gpt_computer_assistant.utils",
        "gpt_computer_assistant.audio",
        "gpt_computer_assistant.audio.tts_providers",
        "gpt_computer_assistant.audio.stt_providers",
    ],
    include_package_data=True,
    install_requires=["requests==2.32.3", "matplotlib==3.9.0"],
    entry_points={
        "console_scripts": ["computerassistant=gpt_computer_assistant.start:start"],
    },
    python_requires=">= 3.9",
    zip_safe=False,
    extras_require={
        "base": install_requires,
        "default": install_requires,
        "agentic": ["crewai==0.30.11"],
        "wakeword": ["pvporcupine", "pyaudio"],
        "api": [
            "flask==3.0.3",
        ],
        "display": [

            "opencv-python",
        ],
        "local_tts": [
            "tensorflow==2.17.0",
            "datasets[audio]==2.20.0",
            "sentencepiece==0.2.0",
            "torch==2.4.0",
            "transformers==4.43.3",
        ],
        "local_stt": [
            "openai-whisper==20231117",
        ],
    },
)





import asyncio
import pathlib
import time
from typing import List, Any, Dict

from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from langchain_mcp import MCPToolkit
from langchain_core.tools import BaseTool




from typing import Any, Dict, List
from langchain_core.tools import BaseTool
from pydantic import Field, PrivateAttr





class MCPToolWrapper(BaseTool):
    """A wrapper for an individual tool managed by the SyncInvocationManager."""
    _manager: Any = PrivateAttr()
    _tool: Any = PrivateAttr()
    
    def __init__(self, tool: BaseTool, manager: "SyncInvocationManager"):
        super().__init__(name=tool.name, description=tool.description)
        self.name = tool.name
        self.description = tool.description
        self._manager = manager
        self._tool = tool

    def _run(self, **kwargs: Any) -> Any:
        """Run the tool synchronously using the SyncInvocationManager."""
        try:
            print(f"Running tool: {self.name} with args: {kwargs}")
            result = self._manager.invoke_tool_sync(self._tool, kwargs)
            if result is None:
                print(f"Tool {self.name} returned no result.")
            else:
                print(f"Tool {self.name} result: {result}")
            return result
        except Exception as e:
            print(f"Error while running tool {self.name}: {e}")
            return None

    async def _arun(self, **kwargs: Any) -> Any:
        """Asynchronous run (if needed), wraps the synchronous call."""
        return self._run(**kwargs)


class MCPToolManager:
    """Manages tools provided by the SyncInvocationManager and converts them into LangChain tools."""
    
    def __init__(self, manager: "SyncInvocationManager"):
        self.manager = manager
        self.tools: List[BaseTool] = []
    
    def load_tools(self) -> List[BaseTool]:
        """Load tools from SyncInvocationManager and wrap them in LangChain-compatible structure."""
        raw_tools = self.manager.get_tools_sync()
     
        self.tools = [MCPToolWrapper(tool, self.manager) for tool in raw_tools]
        return self.tools


class SyncInvocationManager:
    def __init__(self, command: str, args: list[str], env: dict[str, str] | None = None):
        self.loop = asyncio.new_event_loop()
        self.server_params = StdioServerParameters(
            command=command,
            args=args,
            env=env,

        )
        self.client_ctx = None
        self.client = None
        self.session_ctx = None
        self.session = None
        self.toolkit = None
        self._task = None  # Add this line


    async def _start_async(self):
        # Manually enter the stdio_client context
        self.client_ctx = stdio_client(self.server_params)
        self.client = await self.client_ctx.__aenter__()
        read, write = self.client

        # Manually enter the ClientSession context
        self.session_ctx = ClientSession(read, write)
        self.session = await self.session_ctx.__aenter__()

        self.toolkit = MCPToolkit(session=self.session)
        await self.toolkit.initialize()

    def get_tools_sync(self) -> List[BaseTool]:
        # Now that session is open, just return tools directly
        return self.toolkit.get_tools()

    def invoke_tool_sync(self, tool: BaseTool, input_data: Dict[str, Any]) -> Any:
        try:
            return self.loop.run_until_complete(tool.ainvoke(input_data))
        except Exception as e:
            print(f"Error invoking tool {tool.name}: {e}")
            return None

    def start(self):
        asyncio.set_event_loop(self.loop)
        self._task = self.loop.create_task(self._start_async())
        self.loop.run_until_complete(self._task)

    def stop(self):
        if self._task and not self._task.done():
            cleanup_task = self.loop.create_task(self._stop_async())
            self.loop.run_until_complete(cleanup_task)
        self.loop.close()

    async def _stop_async(self):
        # Exit contexts in the same task and loop they were entered
        if self.session_ctx:
            await self.session_ctx.__aexit__(None, None, None)
        if self.client_ctx:
            await self.client_ctx.__aexit__(None, None, None)




def file_system_tool():
    print("""
          
This is file_system_tool

          """)


    manager = SyncInvocationManager(command="npx", args=["-y", "@modelcontextprotocol/server-filesystem", "/"])
    manager.start()
    tool_manager = MCPToolManager(manager)
    tools =  tool_manager.load_tools()
    print(tools)
    return tools


def memory_tool():

    print("""
          
This is memory_tool
          
            """)
    

    manager = SyncInvocationManager(command="npx", args=["-y", "@modelcontextprotocol/server-memory"])
    manager.start()
    tool_manager = MCPToolManager(manager)
    tools =  tool_manager.load_tools()
    print(tools)
    return tools


def playwright():

    print("""
          
This is playwright
          
            """)

    manager = SyncInvocationManager(command="npx", args=["-y", "@executeautomation/playwright-mcp-server"])
    manager.start()
    tool_manager = MCPToolManager(manager)
    tools =  tool_manager.load_tools()
    print(tools)
    return tools


def youtube_transcript():

    print("""
          
This is youtube_transcript
          
            """)

    manager = SyncInvocationManager(command="npx", args=["-y", "@kimtaeyoon83/mcp-server-youtube-transcript"])
    manager.start()
    tool_manager = MCPToolManager(manager)
    tools =  tool_manager.load_tools()
    print(tools)
    return tools

def fetch():

    print("""
          
This is fetch
          
            """)

    manager = SyncInvocationManager(command="uvx", args=["mcp-server-fetch"])
    manager.start()
    tool_manager = MCPToolManager(manager)
    tools =  tool_manager.load_tools()
    print(tools)
    return tools




def websearch():

    print("""
          
This is websearch
          
            """)


    manager = SyncInvocationManager(command="npx", args=["-y", "@mzxrai/mcp-webresearch"])
    manager.start()
    tool_manager = MCPToolManager(manager)
    tools =  tool_manager.load_tools()
    print(tools)
    return tools



custom_mcp_severs_ = []
previous_mcp_servers = []

loaded_mcp_servers = []

def custom_mcp_servers():

    print("Custom MCP Servers")
    global custom_mcp_severs_
    global previous_mcp_servers
    global loaded_mcp_servers
    if custom_mcp_severs_ == previous_mcp_servers:
        print("Returning loaded mcp servers")
        return loaded_mcp_servers
    
    else:
        # The custom_mcp_servers_ list is like [{name: "file_system_tool", command:"npx", args:["-y", "@mzxrai/mcp-webresearch"]}, {name: "memory_tool", command:"npx", args:["-y", "@mzxrai/mcp-webresearch"]}]    
        # We shouldnt load same mcp server twice. For that we need to intersect the custom_mcp_servers_ and previous_mcp_servers
        # and load only the difference
        # This is to avoid loading the same mcp server twice

        # Get the names of the mcp servers that are already loaded
        previous_mcp_server_names = [mcp_server["name"] for mcp_server in loaded_mcp_servers]
        # Get the names of the mcp servers that are in the custom_mcp_servers_ list
        custom_mcp_server_names = [mcp_server["name"] for mcp_server in custom_mcp_severs_]
        # Get the names of the mcp servers that are not loaded
        mcp_server_names_to_load = list(set(custom_mcp_server_names) - set(previous_mcp_server_names))

        # Load the mcp servers that are not loaded

        for mcp_server in custom_mcp_severs_:
            if mcp_server["name"] in mcp_server_names_to_load:
                manager = SyncInvocationManager(command=mcp_server["command"], args=mcp_server["args"])
                manager.start()
                tool_manager = MCPToolManager(manager)
                tools =  tool_manager.load_tools()
                loaded_mcp_servers = loaded_mcp_servers + tools
        previous_mcp_servers = custom_mcp_severs_
        print("Returning loaded mcp servers", loaded_mcp_servers)
        return loaded_mcp_servers
    
def add_custom_mcp_server(name: str, command: str, args: List[str]):
    global custom_mcp_severs_
    print("****************\nAdding custom mcp server")
    print(name, command, args)
    custom_mcp_severs_.append({"name": name, "command": command, "args": args})

def remove_custom_mcp_server(name: str):
    global custom_mcp_severs_
    custom_mcp_severs_ = [mcp_server for mcp_server in custom_mcp_severs_ if mcp_server["name"] != name]

def get_custom_mcp_server(name: str):
    global custom_mcp_severs_
    for mcp_server in custom_mcp_severs_:
        if mcp_server["name"] == name:
            return mcp_server
    return None


            
        

    




the_tools_ = None
def mcp_tools():
    global the_tools_
    if the_tools_ is None:
        #the_tools_ = file_system_tool()
        the_tools_ = []
    return the_tools_ + custom_mcp_servers()
